@@ -1,31 +1,22 @@
 /* cert-object.c - Convert a GPGSM certificate into a PKCS #11 object.
-   Copyright (C) 2006, 2007 g10 Code GmbH
-
-   This file is part of Scute.
- 
-   Scute is free software; you can redistribute it and/or modify it
-   under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
-
-   Scute is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   General Public License for more details.
-
-   You should have received a copy of the GNU General Public License
-   along with Scute; if not, write to the Free Software Foundation,
-   Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
-
-   In addition, as a special exception, g10 Code GmbH gives permission
-   to link this library: with the Mozilla Foundation's code for
-   Mozilla (or with modified versions of it that use the same license
-   as the "Mozilla" code), and distribute the linked executables.  You
-   must obey the GNU General Public License in all respects for all of
-   the code used other than "Mozilla".  If you modify this file, you
-   may extend this exception to your version of the file, but you are
-   not obligated to do so.  If you do not wish to do so, delete this
-   exception statement from your version.  */
+ * Copyright (C) 2006, 2007 g10 Code GmbH
+ *
+ * This file is part of Scute.
+ *
+ * Scute is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License as
+ * published by the Free Software Foundation; either version 2.1 of
+ * the License, or (at your option) any later version.
+ *
+ * Scute is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this program; if not, see <https://gnu.org/licenses/>.
+ * SPDX-License-Identifier: LGPL-2.1-or-later
+ */
 
 #if HAVE_CONFIG_H
 #include <config.h>
@@ -45,21 +36,12 @@
 #include "debug.h"
 
 
-/* Parsing certificates is problematic, as there is no reliable
-   standard.  However, we must parse at least the subject field (see
-   below).  So, disabling cert parsing does currently not disable it
-   completely.  The purpose of the symbol is to document which parts
-   of the code are mandatory, and which are optional.  */
-#define CERT_PARSING 1
-
-
-#if CERT_PARSING || 1
-
 #define atoi_1(p)   (*(p) - '0' )
 #define atoi_2(p)   ((atoi_1(p) * 10) + atoi_1((p)+1))
 #define atoi_4(p)   ((atoi_2(p) * 100) + atoi_2((p)+2))
 
 
+#if 0 /* Currently not used.  */
 static bool
 time_to_ck_date (time_t *atime, CK_DATE *ckdate)
 {
@@ -86,7 +68,7 @@ time_to_ck_date (time_t *atime, CK_DATE *ckdate)
   if (!(broken_time.tm_year >= 0 && broken_time.tm_year <= 8099
 	&& broken_time.tm_mon >= 0 && broken_time.tm_mon <= 11
 	&& broken_time.tm_mday >= 1 && broken_time.tm_mday <= 31))
-    { 
+    {
       DEBUG (DBG_INFO, "unrepresentable time %i-%i-%i",
 	     broken_time.tm_year, broken_time.tm_mon, broken_time.tm_mday);
       return false;
@@ -101,12 +83,12 @@ time_to_ck_date (time_t *atime, CK_DATE *ckdate)
   ckdate->year[1] = LAST_DIGIT (nr);
   nr = nr / 10;
   ckdate->year[0] = LAST_DIGIT (nr);
-  
+
   nr = broken_time.tm_mon + 1;
   ckdate->month[1] = LAST_DIGIT (nr);
   nr = nr / 10;
   ckdate->month[0] = LAST_DIGIT (nr);
-  
+
   nr = broken_time.tm_mday;
   ckdate->day[1] = LAST_DIGIT (nr);
   nr = nr / 10;
@@ -114,7 +96,7 @@ time_to_ck_date (time_t *atime, CK_DATE *ckdate)
 
   return true;
 }
-
+#endif /*0*/
 
 static gpg_error_t
 asn1_get_len (unsigned char **asn1, int *asn1_len, int *rlen)
@@ -241,7 +223,7 @@ asn1_get_element (unsigned char *cert, int cert_len,
   /* We found the subject.  */
   *sub_start = prev_certp;
   *sub_len = certp - prev_certp;
-  
+
   return 0;
 }
 
@@ -380,7 +362,6 @@ asn1_get_public_exp (unsigned char *cert, int cert_len,
 
   return 0;
 }
-#endif
 
 
 static gpg_error_t
@@ -424,7 +405,7 @@ scute_attr_free (CK_ATTRIBUTE_PTR attr, CK_ULONG attr_count)
 
 
 gpg_error_t
-scute_attr_cert (struct cert *cert,
+scute_attr_cert (struct cert *cert, const char *grip,
 		 CK_ATTRIBUTE_PTR *attrp, CK_ULONG *attr_countp)
 {
   CK_RV err = 0;
@@ -442,9 +423,6 @@ scute_attr_cert (struct cert *cert,
   CK_BBOOL obj_token = CK_TRUE;
   CK_BBOOL obj_private = CK_FALSE;
   CK_BBOOL obj_modifiable = CK_FALSE;
-  CK_BYTE obj_label[] = { 'D', 'u', 'm', 'm', 'y', ' ',
-			  'L', 'a', 'b', 'e', 'l' };
-
   CK_CERTIFICATE_TYPE obj_cert_type = CKC_X_509;
   CK_BBOOL obj_trusted = cert->is_trusted;
   CK_ULONG obj_cert_cat = 0;
@@ -453,8 +431,6 @@ scute_attr_cert (struct cert *cert,
   CK_DATE obj_end_date;
   CK_ULONG obj_java_midp_sec_domain = 0;
 
-#if CERT_PARSING || 1
-  /* See below.  */
   err = asn1_get_subject (cert->cert_der, cert->cert_der_len,
 			  &subject_start, &subject_len);
   if (err)
@@ -463,8 +439,7 @@ scute_attr_cert (struct cert *cert,
 	     gpg_strerror (err));
       return err;
     }
-#endif
-#if CERT_PARSING
+
   err = asn1_get_issuer (cert->cert_der, cert->cert_der_len,
 			 &issuer_start, &issuer_len);
   if (err)
@@ -482,7 +457,6 @@ scute_attr_cert (struct cert *cert,
 	     gpg_strerror (err));
       return err;
     }
-#endif
 
 
 #define NR_ATTR_CERT 20
@@ -494,77 +468,109 @@ scute_attr_cert (struct cert *cert,
       return gpg_error (GPG_ERR_ENOMEM);
     }
 
-#define one_attr_ext(type, val, size)					\
-  if (!err)								\
-    err = attr_one (attr, &attr_count, type, val, size)
-
-#define one_attr(type, val) one_attr_ext (type, &val, sizeof (val))
-
-#define empty_attr(type)						\
-  if (!err)								\
-    err = attr_empty (attr, &attr_count, type)
-
-  one_attr (CKA_CLASS, obj_class);
-  one_attr (CKA_TOKEN, obj_token);
-  one_attr (CKA_PRIVATE, obj_private);
-  one_attr (CKA_MODIFIABLE, obj_modifiable);
-  one_attr (CKA_LABEL, obj_label);
-  one_attr (CKA_CERTIFICATE_TYPE, obj_cert_type);
-  one_attr (CKA_TRUSTED, obj_trusted);
-  one_attr (CKA_CERTIFICATE_CATEGORY, obj_cert_cat);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_CLASS,
+                    &obj_class, sizeof obj_class);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_TOKEN,
+                    &obj_token, sizeof obj_token);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_PRIVATE,
+                    &obj_private, sizeof obj_private);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_MODIFIABLE,
+                    &obj_modifiable, sizeof obj_modifiable);
+  if (!err)
+    {
+      if (*cert->certref)
+        err = attr_one (attr, &attr_count, CKA_LABEL,
+                        cert->certref, strlen (cert->certref));
+      else
+        err = attr_one (attr, &attr_count, CKA_LABEL,
+                        "DummyLabel", 10);
+    }
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_CERTIFICATE_TYPE,
+                    &obj_cert_type, sizeof obj_cert_type);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_TRUSTED,
+                    &obj_trusted, sizeof obj_trusted);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_CERTIFICATE_CATEGORY,
+                    &obj_cert_cat, sizeof obj_cert_cat);
 
   /* FIXME: Calculate check_value.  */
-  one_attr (CKA_CHECK_VALUE, obj_check_value);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_CHECK_VALUE,
+                    &obj_check_value, sizeof obj_check_value);
 
 #if 0
   if (time_to_ck_date (&cert->timestamp, &obj_start_date))
     {
-      one_attr (CKA_START_DATE, obj_start_date);
+      if (!err)
+        err = attr_one (attr, &attr_count, CKA_START_DATE,
+                        &obj_start_date, sizeof obj_start_date);
     }
 
   if (time_to_ck_date (&cert->expires, &obj_end_date))
     {
-      one_attr (CKA_END_DATE, obj_end_date);
+      if (!err)
+        err = attr_one (attr, &attr_count, CKA_END_DATE,
+                        &obj_end_date, sizeof obj_end_date);
     }
 #else
   /* For now, we disable these fields.  We can parse them from the
      certificate just as the other data.  However, we would like to
      avoid parsing the certificates at all, let's see how much
      functionality we really need in the PKCS#11 token first.  */
-  empty_attr (CKA_START_DATE);
-  empty_attr (CKA_END_DATE);
+  (void)obj_start_date;
+  (void)obj_end_date;
+  if (!err)
+    err = attr_empty (attr, &attr_count, CKA_START_DATE);
+  if (!err)
+    err = attr_empty (attr, &attr_count, CKA_END_DATE);
 #endif
 
   /* Note: This attribute is mandatory.  Without it, Firefox client
      authentication won't work.  */
-#if CERT_PARSING || 1
-  one_attr_ext (CKA_SUBJECT, subject_start, subject_len);
-#endif
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_SUBJECT,
+                    subject_start, subject_len);
 
-#if 0
-  /* If we get the info directly from the card, we don't have a
-     fingerprint, and parsing the subject key identifier is quite a
-     mouth full.  Let's try a different approach for now.  */
-  one_attr_ext (CKA_ID, cert->fpr, 40);
-#else
-  {
-    char certptr[40];
-    snprintf (certptr, DIM (certptr), "%p", cert);
-    one_attr_ext (CKA_ID, certptr, strlen (certptr));
-  }
-#endif
+  /* We construct the CKA_ID from the CERTREF and the KEYGRIP.  This
+   * allows us to use both values as needed.  */
+  if (!err)
+    {
+      char cka_id_buffer[200];
 
-#if CERT_PARSING
-  one_attr_ext (CKA_ISSUER, issuer_start, issuer_len);
-  one_attr_ext (CKA_SERIAL_NUMBER, serial_start, serial_len);
-#endif
-  one_attr_ext (CKA_VALUE, cert->cert_der, cert->cert_der_len);
-  
-  empty_attr (CKA_URL);
-  empty_attr (CKA_HASH_OF_SUBJECT_PUBLIC_KEY);
-  empty_attr (CKA_HASH_OF_ISSUER_PUBLIC_KEY);
+      snprintf (cka_id_buffer, sizeof cka_id_buffer, "%s %s",
+                *cert->certref ? cert->certref:"-",
+                grip && *grip? grip : "?" );
+      err = attr_one (attr, &attr_count, CKA_ID,
+                      cka_id_buffer, strlen (cka_id_buffer));
+    }
 
-  one_attr (CKA_JAVA_MIDP_SECURITY_DOMAIN, obj_java_midp_sec_domain);
+
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_ISSUER,
+                    issuer_start, issuer_len);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_SERIAL_NUMBER,
+                    serial_start, serial_len);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_VALUE,
+                    cert->cert_der, cert->cert_der_len);
+
+  if (!err)
+    err = attr_empty (attr, &attr_count, CKA_URL);
+  if (!err)
+    err = attr_empty (attr, &attr_count, CKA_HASH_OF_SUBJECT_PUBLIC_KEY);
+  if (!err)
+    err = attr_empty (attr, &attr_count, CKA_HASH_OF_ISSUER_PUBLIC_KEY);
+
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_JAVA_MIDP_SECURITY_DOMAIN,
+                    &obj_java_midp_sec_domain, sizeof obj_java_midp_sec_domain);
 
   if (err)
     {
@@ -584,8 +590,8 @@ scute_attr_cert (struct cert *cert,
 
 
 gpg_error_t
-scute_attr_prv (struct cert *cert, CK_ATTRIBUTE_PTR *attrp,
-		CK_ULONG *attr_countp)
+scute_attr_prv (struct cert *cert, key_info_t kinfo,
+                CK_ATTRIBUTE_PTR *attrp, CK_ULONG *attr_countp)
 {
   CK_RV err = 0;
   CK_ATTRIBUTE_PTR attr;
@@ -602,9 +608,6 @@ scute_attr_prv (struct cert *cert, CK_ATTRIBUTE_PTR *attrp,
   CK_BBOOL obj_token = CK_TRUE;
   CK_BBOOL obj_private = CK_FALSE;
   CK_BBOOL obj_modifiable = CK_FALSE;
-  CK_BYTE obj_label[] = { 'O', 'P', 'E', 'N', 'P', 'G',
-			  'P', '.', '3' };
-
   CK_KEY_TYPE obj_key_type = CKK_RSA;
   CK_DATE obj_start_date;
   CK_DATE obj_end_date;
@@ -614,8 +617,8 @@ scute_attr_prv (struct cert *cert, CK_ATTRIBUTE_PTR *attrp,
   CK_MECHANISM_TYPE obj_mechanisms[] = { CKM_RSA_PKCS };
 
   CK_BBOOL obj_sensitive = CK_TRUE;
-  CK_BBOOL obj_decrypt = CK_FALSE;	/* Authentication only for now.  */
-  CK_BBOOL obj_sign = CK_TRUE;
+  CK_BBOOL obj_decrypt = CK_FALSE;      /* Updated below.  */
+  CK_BBOOL obj_sign = CK_FALSE;         /* Updated below.  */
   CK_BBOOL obj_sign_recover = CK_FALSE;
   CK_BBOOL obj_unwrap = CK_FALSE;
   CK_BBOOL obj_extractable = CK_FALSE;
@@ -624,7 +627,13 @@ scute_attr_prv (struct cert *cert, CK_ATTRIBUTE_PTR *attrp,
   CK_BBOOL obj_wrap_with_trusted = CK_FALSE;
   CK_BBOOL obj_always_authenticate = CK_FALSE;
 
-#if CERT_PARSING
+  if (kinfo->usage.sign || kinfo->usage.cert || kinfo->usage.auth)
+    obj_sign = CK_TRUE;
+
+  if (kinfo->usage.encr)
+    obj_decrypt = CK_TRUE;
+
+
   err = asn1_get_subject (cert->cert_der, cert->cert_der_len,
 			  &subject_start, &subject_len);
   if (err)
@@ -649,7 +658,6 @@ scute_attr_prv (struct cert *cert, CK_ATTRIBUTE_PTR *attrp,
 	     gpg_strerror (err));
       return err;
     }
-#endif
 
 #define NR_ATTR_PRV 27
   attr = malloc (sizeof (CK_ATTRIBUTE) * NR_ATTR_PRV);
@@ -660,38 +668,44 @@ scute_attr_prv (struct cert *cert, CK_ATTRIBUTE_PTR *attrp,
       return gpg_error (GPG_ERR_ENOMEM);
     }
 
-#undef one_attr_ext
-#define one_attr_ext(type, val, size)					\
-  if (!err)								\
-    err = attr_one (attr, &attr_count, type, val, size)
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_CLASS,
+                    &obj_class, sizeof obj_class);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_TOKEN,
+                    &obj_token, sizeof obj_token);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_PRIVATE,
+                    &obj_private, sizeof obj_private);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_MODIFIABLE,
+                    &obj_modifiable, sizeof obj_modifiable);
+  if (!err)
+    {
+      if (*cert->certref)
+        err = attr_one (attr, &attr_count, CKA_LABEL,
+                        cert->certref, strlen (cert->certref));
+      else
+        err = attr_one (attr, &attr_count, CKA_LABEL,
+                        "DummyLabel", 10);
+    }
 
-#undef one_attr
-#define one_attr(type, val) one_attr_ext (type, &val, sizeof (val))
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_KEY_TYPE,
+                    &obj_key_type, sizeof obj_key_type);
 
-#undef empty_attr
-#define empty_attr(type)						\
-  if (!err)								\
-    err = attr_empty (attr, &attr_count, type)
+  /* We construct the CKA_ID from the CERTREF and the KEYGRIP.  This
+   * allows us to use both values as needed.  */
+  if (!err)
+    {
+      char cka_id_buffer[200];
 
-  one_attr (CKA_CLASS, obj_class);
-  one_attr (CKA_TOKEN, obj_token);
-  one_attr (CKA_PRIVATE, obj_private);
-  one_attr (CKA_MODIFIABLE, obj_modifiable);
-  one_attr (CKA_LABEL, obj_label);
-
-  one_attr (CKA_KEY_TYPE, obj_key_type);
-#if 0
-  /* If we get the info directly from the card, we don't have a
-     fingerprint, and parsing the subject key identifier is quite a
-     mouth full.  Let's try a different approach for now.  */
-  one_attr_ext (CKA_ID, cert->fpr, 40);
-#else
-  {
-    char certptr[40];
-    snprintf (certptr, DIM (certptr), "%p", cert);
-    one_attr_ext (CKA_ID, certptr, strlen (certptr));
-  }
-#endif
+      snprintf (cka_id_buffer, sizeof cka_id_buffer, "%s %s",
+                *cert->certref ? cert->certref:"-",
+                kinfo->grip && *kinfo->grip? kinfo->grip : "?" );
+      err = attr_one (attr, &attr_count, CKA_ID,
+                      cka_id_buffer, strlen (cka_id_buffer));
+    }
 
 #if 0
   /* For now, we disable these fields.  We can parse them from the
@@ -703,47 +717,86 @@ scute_attr_prv (struct cert *cert, CK_ATTRIBUTE_PTR *attrp,
      gpgsm.  */
   if (time_to_ck_date (&cert->timestamp, &obj_start_date))
     {
-      one_attr (CKA_START_DATE, obj_start_date);
+      if (!err)
+        err = attr_one (attr, &attr_count, CKA_START_DATE,
+                        &obj_start_date, sizeof obj_start_date);
     }
 
   if (time_to_ck_date (&cert->expires, &obj_end_date))
     {
-      one_attr (CKA_END_DATE, obj_end_date);
+      if (!err)
+        err = attr_one (attr, &attr_count, CKA_END_DATE,
+                        &obj_end_date, sizeof obj_end_date);
     }
 #else
   /* For now, we disable these fields.  We can parse them from the
      certificate just as the other data.  However, we would like to
      avoid parsing the certificates at all, let's see how much
      functionality we really need in the PKCS#11 token first.  */
-  empty_attr (CKA_START_DATE);
-  empty_attr (CKA_END_DATE);
+  (void)obj_start_date;
+  (void)obj_end_date;
+  if (!err)
+    err = attr_empty (attr, &attr_count, CKA_START_DATE);
+  if (!err)
+    err = attr_empty (attr, &attr_count, CKA_END_DATE);
 #endif
 
-  one_attr (CKA_DERIVE, obj_derive);
-  one_attr (CKA_LOCAL, obj_local);
-  one_attr (CKA_KEY_GEN_MECHANISM, obj_key_gen);
-  one_attr (CKA_ALLOWED_MECHANISMS, obj_mechanisms);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_DERIVE,
+                    &obj_derive, sizeof obj_derive);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_LOCAL,
+                    &obj_local, sizeof obj_local);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_KEY_GEN_MECHANISM,
+                    &obj_key_gen, sizeof obj_key_gen);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_ALLOWED_MECHANISMS,
+                    &obj_mechanisms, sizeof obj_mechanisms);
 
-#if CERT_PARSING
-  one_attr_ext (CKA_SUBJECT, subject_start, subject_len);
-#endif
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_SUBJECT,
+                    subject_start, subject_len);
 
-  one_attr (CKA_SENSITIVE, obj_sensitive);
-  one_attr (CKA_DECRYPT, obj_decrypt);
-  one_attr (CKA_SIGN, obj_sign);
-  one_attr (CKA_SIGN_RECOVER, obj_sign_recover);
-  one_attr (CKA_UNWRAP, obj_unwrap);
-  one_attr (CKA_EXTRACTABLE, obj_extractable);
-  one_attr (CKA_ALWAYS_SENSITIVE, obj_always_sensitive);
-  one_attr (CKA_NEVER_EXTRACTABLE, obj_never_extractable);
-  one_attr (CKA_WRAP_WITH_TRUSTED, obj_wrap_with_trusted);
-  empty_attr (CKA_UNWRAP_TEMPLATE);
-  one_attr (CKA_ALWAYS_AUTHENTICATE, obj_always_authenticate);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_SENSITIVE,
+                    &obj_sensitive, sizeof obj_sensitive);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_DECRYPT,
+                    &obj_decrypt, sizeof obj_decrypt);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_SIGN,
+                    &obj_sign, sizeof obj_sign);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_SIGN_RECOVER,
+                    &obj_sign_recover, sizeof obj_sign_recover);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_UNWRAP,
+                    &obj_unwrap, sizeof obj_unwrap);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_EXTRACTABLE,
+                    &obj_extractable, sizeof obj_extractable);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_ALWAYS_SENSITIVE,
+                    &obj_always_sensitive, sizeof obj_always_sensitive);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_NEVER_EXTRACTABLE,
+                    &obj_never_extractable, sizeof obj_never_extractable);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_WRAP_WITH_TRUSTED,
+                    &obj_wrap_with_trusted, sizeof obj_wrap_with_trusted);
+  if (!err)
+    err = attr_empty (attr, &attr_count, CKA_UNWRAP_TEMPLATE);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_ALWAYS_AUTHENTICATE,
+                    &obj_always_authenticate, sizeof obj_always_authenticate);
 
-#if CERT_PARSING
-  one_attr_ext (CKA_MODULUS, modulus_start, modulus_len);
-  one_attr_ext (CKA_PUBLIC_EXPONENT, public_exp_start, public_exp_len);
-#endif
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_MODULUS,
+                    modulus_start, modulus_len);
+  if (!err)
+    err = attr_one (attr, &attr_count, CKA_PUBLIC_EXPONENT,
+                    public_exp_start, public_exp_len);
 
   if (err)
     {
