@@ -66,7 +66,7 @@ void scute_slots_finalize (void);
    Mozilla NSS currently assumes that the slot list never shrinks (see
    TODO file for a discussion).  This is the only function allowed to
    manipulate the slot list.  */
-CK_RV slots_update (void);
+CK_RV slots_update_all (void);
 
 /* Update the slot SLOT.  */
 CK_RV slots_update_slot (slot_iterator_t id);
@@ -201,7 +201,8 @@ CK_RV session_get_search_result (slot_iterator_t id, session_iterator_t sid,
 
 /* The core of C_SignInit.  */
 CK_RV session_set_signing_key (slot_iterator_t id, session_iterator_t sid,
-			       object_iterator_t key);
+			       object_iterator_t key,
+                               CK_MECHANISM_TYPE mechanism_type);
 
 /* The core of C_Sign.  */
 CK_RV session_sign (slot_iterator_t id, session_iterator_t sid,
